
/* Chrysalide - Outil d'analyse de fichiers binaires
 * symbol.c - équivalent Python du fichier "format/symbol.h"
 *
 * Copyright (C) 2015 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "symbol.h"


#include <pygobject.h>


#include <i18n.h>


#include <format/symbol.h>


#include "../helpers.h"
#include "../analysis/routine.h"
#include "../analysis/db/items/comment.h"
#include "../arch/instruction.h"
#include "../arch/vmpa.h"



/* Effectue une comparaison avec un objet Python 'BinSymbol'. */
static PyObject *py_binary_symbol_richcompare(PyObject *, PyObject *, int);

/* Définit un autre nom pour le symbole. */
static PyObject *py_binary_symbol_set_alt_label(PyObject *, PyObject *);

/* Raffine la définition de l'emplacement d'un symbole. */
static PyObject *py_binary_symbol_fix_range(PyObject *, PyObject *);

/* Attache la routine associée au symbole. */
static PyObject *py_binary_symbol_attach_routine(PyObject *, PyObject *);

/* Attache l'instruction associée au symbole. */
static PyObject *py_binary_symbol_attach_instruction(PyObject *, PyObject *);

/* Fournit le type du symbole. */
static PyObject *py_binary_symbol_get_target_type(PyObject *, void *);

/* Fournit un étiquette pour viser un symbole. */
static PyObject *py_binary_symbol_get_label(PyObject *, void *);

/* Fournit l'emplacement où se situe un symbole. */
static PyObject *py_binary_symbol_get_range(PyObject *, void *);

/* Fournit l'éventuelle routine associée au symbole. */
static PyObject *py_binary_symbol_get_routine(PyObject *, void *);

/* Fournit l'éventuelle instruction associée au symbole. */
static PyObject *py_binary_symbol_get_instruction(PyObject *, void *);

/* Fournit l'éventuel commentaire associé au symbole. */
static PyObject *py_binary_symbol_get_comment(PyObject *, void *);

/* Ajoute un commentaire facultatif au symbole. */
static int py_binary_symbol_set_comment(PyObject *, PyObject *, void *);

/* Définit les constantes pour les symboles binaires. */
static bool py_binary_symbol_define_constants(PyTypeObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : obj  = objet Python à tenter de convertir.                   *
*                addr = structure équivalente pour Chrysalide.                *
*                                                                             *
*  Description : Effectue une comparaison avec un objet Python 'BinSymbol'.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_richcompare(PyObject *a, PyObject *b, int op)
{
    PyObject *result;                       /* Chaîne à retourner          */
    int ret;                                /* Bilan de lecture des args.  */
    const GBinSymbol *sym_a;                /* Premier élément à traiter   */
    const GBinSymbol *sym_b;                /* Second élément à traiter    */
    int status;                             /* Résultat d'une comparaison  */

    ret = PyObject_IsInstance(b, (PyObject *)get_python_binary_symbol_type());
    if (!ret)
    {
        result = Py_NotImplemented;
        goto cmp_done;
    }

    sym_a = G_BIN_SYMBOL(pygobject_get(a));
    sym_a = G_BIN_SYMBOL(pygobject_get(b));

    status = g_binary_symbol_cmp(&sym_a, &sym_b);

    switch (op)
    {
        case Py_LT:
            result = status < 0 ? Py_True : Py_False;
            break;

        case Py_LE:
            result = status <= 0 ? Py_True : Py_False;
            break;

        case Py_EQ:
            result = status == 0 ? Py_True : Py_False;
            break;

        case Py_NE:
            result = status != 0 ? Py_True : Py_False;
            break;

        case Py_GT:
            result = status > 0 ? Py_True : Py_False;
            break;

        case Py_GE:
            result = status >= 0 ? Py_True : Py_False;
            break;

        default:
            result = Py_NotImplemented;
            break;

    }

 cmp_done:

    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : type = type de l'objet à instancier.                         *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Crée un nouvel objet Python de type 'DbComment'.             *
*                                                                             *
*  Retour      : Instance Python mise en place.                               *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    PyObject *result;                       /* Bilan à retourner           */
    SymbolType stype;                       /* Type prévu pour le  symbole */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Version GLib du symble      */

    ret = PyArg_ParseTuple(args, "l", &stype);
    if (!ret) return NULL;

    if (stype >= STP_COUNT)
    {
        PyErr_SetString(PyExc_ValueError, _("Invalid type of message"));
        return NULL;
    }

    symbol = g_binary_symbol_new(stype);

    result = pygobject_new(G_OBJECT(symbol));
    g_object_unref(symbol);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Définit un autre nom pour le symbole.                        *
*                                                                             *
*  Retour      : None.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_set_alt_label(PyObject *self, PyObject *args)
{
    const char *alt;                        /* Etiquette alternative       */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Elément à consulter         */

    ret = PyArg_ParseTuple(args, "s", &alt);
    if (!ret) return NULL;

    symbol = G_BIN_SYMBOL(pygobject_get(self));

    g_binary_symbol_set_alt_label(symbol, alt);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Raffine la définition de l'emplacement d'un symbole.         *
*                                                                             *
*  Retour      : None.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_fix_range(PyObject *self, PyObject *args)
{
    PyObject *py_vmpa;                      /* Localisation version Python */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Elément à consulter         */

    ret = PyArg_ParseTuple(args, "O", &py_vmpa);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(py_vmpa, (PyObject *)get_python_vmpa_type());
    if (!ret) return NULL;

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    g_binary_symbol_fix_range(symbol, get_internal_vmpa(py_vmpa));

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Attache la routine associée au symbole.                      *
*                                                                             *
*  Retour      : None.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_attach_routine(PyObject *self, PyObject *args)
{
    PyObject *py_routine;                   /* Routine version Python      */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    GBinRoutine *routine;                   /* Routine à attacher          */

    ret = PyArg_ParseTuple(args, "O", &py_routine);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(py_routine, (PyObject *)get_python_binary_routine_type());
    if (!ret) return NULL;

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    routine = G_BIN_ROUTINE(pygobject_get(py_routine));

    g_binary_symbol_attach_routine(symbol, routine);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Attache l'instruction associée au symbole.                   *
*                                                                             *
*  Retour      : None.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_attach_instruction(PyObject *self, PyObject *args)
{
    PyObject *py_instr;                     /* Instruction version Python  */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    GArchInstruction *instr;                /* Instruction à attacher      */

    ret = PyArg_ParseTuple(args, "O", &py_instr);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(py_instr, (PyObject *)get_python_arch_instruction_type());
    if (!ret) return NULL;

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    instr = G_ARCH_INSTRUCTION(pygobject_get(py_instr));

    g_binary_symbol_attach_instruction(symbol, instr);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit le type du symbole.                                  *
*                                                                             *
*  Retour      : Type de symbole représenté.                                  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_target_type(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    SymbolType type;                        /* Type de symbole représenté  */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    type = g_binary_symbol_get_target_type(symbol);

    result = PyLong_FromLong(type);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit un étiquette pour viser un symbole.                  *
*                                                                             *
*  Retour      : Chaîne de caractères renvoyant au symbole.                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_label(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    const char *label;                      /* Désignation courante        */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    label = g_binary_symbol_get_label(symbol);

    result = PyUnicode_FromString(label);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit l'emplacement où se situe un symbole.                *
*                                                                             *
*  Retour      : Zone mémoire couverte par le symbole.                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_range(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    const mrange_t *range;                  /* Couverture courante         */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    range = g_binary_symbol_get_range(symbol);

    result = build_from_internal_mrange(range);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit l'éventuelle routine associée au symbole.            *
*                                                                             *
*  Retour      : Instance en place ou None si aucune.                         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_routine(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    GBinRoutine *routine;                   /* Routine attachée            */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    routine = g_binary_symbol_get_routine(symbol);

    result = pygobject_new(G_OBJECT(routine));

    return result;

}



/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit l'éventuelle instruction associée au symbole.        *
*                                                                             *
*  Retour      : Instance en place ou None si aucune.                         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_instruction(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    GArchInstruction *instr;                /* Routine attachée            */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    instr = g_binary_symbol_get_instruction(symbol);

    result = pygobject_new(G_OBJECT(instr));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit l'éventuel commentaire associé au symbole.           *
*                                                                             *
*  Retour      : Instance Python du commentaire GLib ou None si aucun.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_comment(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    GDbComment *comment;                    /* Commentaire à associé       */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    comment = g_binary_symbol_get_comment(symbol);

    if (comment != NULL)
        result = pygobject_new(G_OBJECT(comment));
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                value   = valeur fournie à intégrer ou prendre en compte.    *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Ajoute un commentaire facultatif au symbole.                 *
*                                                                             *
*  Retour      : Bilan de l'opération pour Python.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_binary_symbol_set_comment(PyObject *self, PyObject *value, void *closure)
{
    GBinSymbol *symbol;                     /* Elément à consulter         */
    int ret;                                /* Bilan de lecture des args.  */
    GDbComment *comment;                    /* Commentaire à associer      */

    symbol = G_BIN_SYMBOL(pygobject_get(self));

    if (value == Py_None)
        g_binary_symbol_set_comment(symbol, NULL);

    else
    {
        ret = PyObject_IsInstance(value, (PyObject *)get_python_db_comment_type());
        if (!ret) return -1;

        comment = G_DB_COMMENT(pygobject_get(value));

        g_binary_symbol_set_comment(symbol, comment);

    }

    return 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : obj_type = type dont le dictionnaire est à compléter.        *
*                                                                             *
*  Description : Définit les constantes pour les symboles binaires.           *
*                                                                             *
*  Retour      : true en cas de succès de l'opération, false sinon.           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool py_binary_symbol_define_constants(PyTypeObject *obj_type)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    result &= PyDict_AddIntMacro(obj_type, STP_DATA);
    result &= PyDict_AddIntMacro(obj_type, STP_ROUTINE);
    result &= PyDict_AddIntMacro(obj_type, STP_CODE_LABEL);
    result &= PyDict_AddIntMacro(obj_type, STP_OBJECT);
    result &= PyDict_AddIntMacro(obj_type, STP_FUNCTION);
    result &= PyDict_AddIntMacro(obj_type, STP_ENTRY_POINT);
    result &= PyDict_AddIntMacro(obj_type, STP_STRING);
    result &= PyDict_AddIntMacro(obj_type, STP_RO_STRING);
    result &= PyDict_AddIntMacro(obj_type, STP_COUNT);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_binary_symbol_type(void)
{
    static PyMethodDef py_bin_symbol_methods[] = {
        {
            "set_alt_label", py_binary_symbol_set_alt_label,
            METH_VARARGS,
            "set_alt_label($self, alt, /)\n--\n\nSet an alternative label for the symbol."
        },
        {
            "fix_range", py_binary_symbol_fix_range,
            METH_VARARGS,
            "fix_range($self, range, /)\n--\n\nRefine the location of the current symbol."
        },
        {
            "attach_routine", py_binary_symbol_attach_routine,
            METH_VARARGS,
            "attach_routine($self, instr, /)\n--\n\nAttach a provided routine to the current symbol."
        },
        {
            "attach_instruction", py_binary_symbol_attach_instruction,
            METH_VARARGS,
            "attach_instruction($self, instr, /)\n--\n\nAttach a provided instruction to the current symbol."
        },
        { NULL }
    };

    static PyGetSetDef py_bin_symbol_getseters[] = {
        {
            "target_type", py_binary_symbol_get_target_type, NULL,
            "Type of the current symbol.", NULL
        },
        {
            "label", py_binary_symbol_get_label, NULL,
            "Label of the symbol, provided by the internal component or by an alternative label.", NULL
        },
        {
            "range", py_binary_symbol_get_range, NULL,
            "Range covered by the symbol.", NULL
        },
        {
            "routine", py_binary_symbol_get_routine, NULL,
            "Potential routine attached to the symbol.", NULL
        },
        {
            "instruction", py_binary_symbol_get_instruction, NULL,
            "Potential instruction attached to the symbol.", NULL
        },
        {
            "comment", py_binary_symbol_get_comment, py_binary_symbol_set_comment,
            "Optional comment linked to the symbol.", NULL
        },
        { NULL }
    };

    static PyTypeObject py_bin_symbol_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.format.BinSymbol",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT,

        .tp_doc         = "PyChrysalide binary symbol",

        .tp_richcompare = py_binary_symbol_richcompare,

        .tp_methods     = py_bin_symbol_methods,
        .tp_getset      = py_bin_symbol_getseters,
        .tp_new         = (newfunc)py_binary_symbol_new

    };

    return &py_bin_symbol_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.format.BinSymbol'.     *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_binary_symbol(PyObject *module)
{
    PyTypeObject *py_bin_symbol_type;       /* Type Python 'BinSymbol'     */
    int ret;                                /* Bilan d'un appel            */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_bin_symbol_type = get_python_binary_symbol_type();

    py_bin_symbol_type->tp_base = &PyGObject_Type;
    py_bin_symbol_type->tp_basicsize = py_bin_symbol_type->tp_base->tp_basicsize;

    if (PyType_Ready(py_bin_symbol_type) != 0)
        return false;

    if (!py_binary_symbol_define_constants(py_bin_symbol_type))
        return false;

    Py_INCREF(py_bin_symbol_type);
    ret = PyModule_AddObject(module, "BinSymbol", (PyObject *)py_bin_symbol_type);
    if (ret != 0) return false;

    dict = PyModule_GetDict(module);
    pygobject_register_class(dict, "BinSymbol", G_TYPE_BIN_SYMBOL, py_bin_symbol_type,
                             Py_BuildValue("(O)", py_bin_symbol_type->tp_base));

    return true;

}
